@@ -644,7 +644,7 @@ get_plugin_loader (PeasEngine     *engine,
 
   if (loader_info->module == NULL)
     {
-      g_hash_table_insert (loaders, info->loader, NULL);
+      g_hash_table_insert (loaders, g_strdup (info->loader), NULL);
       return NULL;
     }
 
@@ -665,7 +665,7 @@ get_plugin_loader (PeasEngine     *engine,
         g_object_unref (loader_info->loader);
 
       g_object_unref (loader_info->module);
-      g_hash_table_insert (loaders, info->loader, NULL);
+      g_hash_table_insert (loaders, g_strdup (info->loader), NULL);
       return NULL;
     }
 
@@ -777,7 +777,8 @@ load_plugin (PeasEngine     *engine,
       dep_info = peas_engine_get_plugin_info (engine, dependencies[i]);
       if (!dep_info)
         {
-          g_warning ("Plugin not found: %s", dependencies[i]);
+          g_warning ("Could not find plugin '%s' for plugin '%s'",
+                     dependencies[i], peas_plugin_info_get_module_name (info));
           g_set_error (&info->error,
                        PEAS_PLUGIN_INFO_ERROR,
                        PEAS_PLUGIN_INFO_ERROR_DEP_NOT_FOUND,
@@ -792,7 +793,7 @@ load_plugin (PeasEngine     *engine,
                        PEAS_PLUGIN_INFO_ERROR,
                        PEAS_PLUGIN_INFO_ERROR_LOADING_FAILED,
                        _("Dependency '%s' failed to load"),
-                       dep_info->name);
+                       peas_plugin_info_get_name (dep_info));
           goto error;
         }
     }
@@ -802,7 +803,7 @@ load_plugin (PeasEngine     *engine,
   if (loader == NULL)
     {
       g_warning ("Could not find loader '%s' for plugin '%s'",
-                 info->loader, info->name);
+                 info->loader, peas_plugin_info_get_module_name (info));
       g_set_error (&info->error,
                    PEAS_PLUGIN_INFO_ERROR,
                    PEAS_PLUGIN_INFO_ERROR_LOADER_NOT_FOUND,
@@ -813,7 +814,8 @@ load_plugin (PeasEngine     *engine,
 
   if (!peas_plugin_loader_load (loader, info))
     {
-      g_warning ("Error loading plugin '%s'", info->name);
+      g_warning ("Error loading plugin '%s'",
+                 peas_plugin_info_get_module_name (info));
       g_set_error (&info->error,
                    PEAS_PLUGIN_INFO_ERROR,
                    PEAS_PLUGIN_INFO_ERROR_LOADING_FAILED,
@@ -821,6 +823,8 @@ load_plugin (PeasEngine     *engine,
       goto error;
     }
 
+  g_debug ("Loaded plugin '%s'", peas_plugin_info_get_module_name (info));
+
   return TRUE;
 
 error:
@@ -903,6 +907,8 @@ peas_engine_unload_plugin_real (PeasEngine     *engine,
   peas_plugin_loader_garbage_collect (loader);
   peas_plugin_loader_unload (loader, info);
 
+  g_debug ("Unloaded plugin '%s'", peas_plugin_info_get_module_name (info));
+
   if (!engine->priv->in_dispose)
     g_object_notify (G_OBJECT (engine), "loaded-plugins");
 }
@@ -942,6 +948,7 @@ peas_engine_provides_extension (PeasEngine     *engine,
 
   g_return_val_if_fail (PEAS_IS_ENGINE (engine), FALSE);
   g_return_val_if_fail (info != NULL, FALSE);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (extension_type), FALSE);
 
   if (!peas_plugin_info_is_loaded (info))
     return FALSE;
@@ -978,14 +985,26 @@ peas_engine_create_extensionv (PeasEngine     *engine,
                                GParameter     *parameters)
 {
   PeasPluginLoader *loader;
+  PeasExtension *extension;
 
   g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
   g_return_val_if_fail (info != NULL, NULL);
   g_return_val_if_fail (peas_plugin_info_is_loaded (info), NULL);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (extension_type), FALSE);
 
   loader = get_plugin_loader (engine, info);
-  return peas_plugin_loader_create_extension (loader, info, extension_type,
-                                              n_parameters, parameters);
+  extension = peas_plugin_loader_create_extension (loader, info, extension_type,
+                                                   n_parameters, parameters);
+
+  if (!PEAS_IS_EXTENSION (extension))
+    {
+      g_warning ("Plugin '%s' does not provide a '%s' extension",
+                 peas_plugin_info_get_module_name (info),
+                 g_type_name (extension_type));
+      return NULL;
+    }
+
+  return extension;
 }
 
 /**
@@ -1018,12 +1037,19 @@ peas_engine_create_extension_valist (PeasEngine     *engine,
   GParameter *parameters;
   PeasExtension *exten;
 
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
+  g_return_val_if_fail (info != NULL, NULL);
+  g_return_val_if_fail (peas_plugin_info_is_loaded (info), NULL);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (extension_type), FALSE);
+
   type_struct = _g_type_struct_ref (extension_type);
 
   if (!_valist_to_parameter_list (extension_type, type_struct, first_property,
                                   var_args, &parameters, &n_parameters))
     {
-      g_return_val_if_reached (NULL);
+      /* Already warned */
+      _g_type_struct_unref (extension_type, type_struct);
+      return NULL;
     }
 
   exten = peas_engine_create_extensionv (engine, info, extension_type,
@@ -1073,6 +1099,11 @@ peas_engine_create_extension (PeasEngine     *engine,
   va_list var_args;
   PeasExtension *exten;
 
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
+  g_return_val_if_fail (info != NULL, NULL);
+  g_return_val_if_fail (peas_plugin_info_is_loaded (info), NULL);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (extension_type), FALSE);
+
   va_start (var_args, first_property);
   exten = peas_engine_create_extension_valist (engine, info, extension_type,
                                                first_property, var_args);
