@@ -29,7 +29,6 @@ struct _PeasPluginInfo {
   /*< private >*/
   gint refcount;
 
-  gchar *file;
   gchar *module_dir;
   gchar *data_dir;
 
@@ -45,16 +44,17 @@ struct _PeasPluginInfo {
   gchar *website;
   gchar *version;
   gchar *help_uri;
-  guint iage;
-  GHashTable *keys;
 
-  gint loaded : 1;
+  GError *error;
+
+  guint loaded : 1;
   /* A plugin is unavailable if it is not possible to load it
      due to an error loading the plugin module (e.g. for Python plugins
      when the interpreter has not been correctly initializated) */
-  gint available : 1;
+  guint available : 1;
 
   guint builtin : 1;
+  guint hidden : 1;
 };
 
 PeasPluginInfo *_peas_plugin_info_new   (const gchar    *filename,
